zero_one_to_ab = function(x, a, b){
    ##
    ## Maps points in [0, 1] -> [a, b]
    ##
    a + (b-a)*x
}

trisection_search = function(a, b, opt_fn, xtol=1.e-6, max_iters=10){
    ##
    ## Returns an interval containing the maximum of a unimodal function
    ##
    CI = c(a, b) ## the current interval

    ## Do one step 'by hand':
    ##
    x_points = CI[1] + c(1/4, 1/2, 3/4)*(CI[2]-CI[1]) ## where we have sampled our function
    y_points = c(opt_fn(x_points[1]), opt_fn(x_points[2]), opt_fn(x_points[3])) ## the value of our function at the sample points
    n_iters = 1
    done = FALSE
    while( !done ){
        range = CI[2]-CI[1]
        x_opt = mean(CI)
        print(sprintf('n_iters= %5d; CI= [%.6f, %.6f]; range= %.4f; x_opt= %.6f', n_iters, CI[1], CI[2], range, x_opt))
        if( range < xtol || n_iters >= max_iters ){
            break
        }

        ## Look at the location of the maximum location, adjust our current interval, and reuse computations:
        ##
        max_indx = which.max(y_points)
        if( max_indx==1 ){
            CI = c(CI[1], x_points[2])
            x_points = c(NA, x_points[1], NA)
            y_points = c(NA, y_points[1], NA)
        }else if( max_indx==2 ){
            CI = c(x_points[1], x_points[3])
            x_points = c(NA, x_points[2], NA)
            y_points = c(NA, y_points[2], NA)
        }else{
            CI = c(x_points[2], CI[2])
            x_points = c(NA, x_points[3], NA)
            y_points = c(NA, y_points[3], NA)
        }

        ## Compute the missing (x, y) values:
        ##
        x_points[1] = zero_one_to_ab(1/4, CI[1], CI[2])
        y_points[1] = opt_fn(x_points[1])
        x_points[3] = zero_one_to_ab(3/4, CI[1], CI[2])
        y_points[3] = opt_fn(x_points[3])

        n_iters = n_iters + 1
    }

    return(CI)
}

golden_section_search = function(a, b, opt_fn, xtol=1.e-6, max_iters=10){
    ##
    ## Returns an interval containing the maximum of a unimodal function
    ##
    CI = c(a, b) ## the current interval

    x_2 = (sqrt(5) - 1)/2 ## 0.618034
    x_1 = 1 - x_2 ## 0.381966

    ## Do one step 'by hand':
    ##
    x_points = CI[1] + c(x_1, x_2)*(CI[2]-CI[1]) ## where we have sampled our function
    y_points = c(opt_fn(x_points[1]), opt_fn(x_points[2])) ## the value of our function at the sample points
    n_iters = 1
    done = FALSE
    while( !done ){
        range = CI[2]-CI[1]
        x_opt = mean(CI)
        relative_range = x_2^(n_iters-1) # true range is c(a, b)*relative_range
        print(sprintf('n_iters= %5d; CI= [%.6f, %.6f]; range= %.4f; x_opt= %.6f', n_iters, CI[1], CI[2], range, x_opt))
        if( range < xtol || n_iters >= max_iters ){
            break
        }

        ## Look at the location of the maximum location and adjust our current interval:
        ##
        max_indx = which.max(y_points)
        if( max_indx==1 ){
            CI = c(CI[1], x_points[2])
        }else{
            CI = c(x_points[1], CI[2])
        }
        x_points = CI[1] + c(x_1, x_2)*(CI[2]-CI[1])
        y_points = c(opt_fn(x_points[1]), opt_fn(x_points[2])) ## the value of our function at the sample points

        n_iters = n_iters + 1
    }

    return(CI)
}


